(** Time-stamp: <modified the 01/02/2016 (at 17:12) by Erwan Jahier> *)

(** topological sort of actions (that may optimize test openning) *)

open ActionsDeps

let info msg =
  let t = Sys.time() in
  Lv6Verbose.exe ~level:1 (fun () -> Printf.eprintf "%4.4f: %s%!" t msg; flush stderr)


(*********************************************************************************)
module OrderedAction = struct
  type t = Action.t
  let compare = compare
end
module MapAction = Map.Make(OrderedAction)


type color = Grey | Black (* in process | done *)
type color_table = color MapAction.t

exception DependencyCycle of Action.t * Action.t list
(* exception DependencyCycle of Soc.var_expr list *)

let (grey_actions : color_table -> Action.t list) =
  fun ct -> 
    MapAction.fold
      (fun action color acc -> if color=Grey then action::acc else acc) ct []

let rec (visit : ActionsDeps.t -> color_table -> Action.t -> color_table) =
  fun succ_t color_t n ->
    if not (ActionsDeps.have_deps succ_t n) then MapAction.add n Black color_t else
      let color_t =
	     List.fold_left
	       (fun color_t nt -> 
	         try
	           match MapAction.find nt color_t with
		          | Grey -> raise (DependencyCycle (n, grey_actions color_t))
		          | Black -> color_t
	         with 
		          (* The node [nt] is white *)
		          Not_found -> visit succ_t color_t nt
	       ) 
	       (MapAction.add n Grey color_t)
	       (ActionsDeps.find_deps succ_t n)
      in
	   MapAction.add n Black color_t

(* TEDLT *)
let (check_there_is_no_cycle : Action.t list -> ActionsDeps.t -> unit) =
  fun actions t ->
  ignore (List.fold_left
            (fun acc action -> visit t acc action) MapAction.empty actions)

let (topo_sort : Action.t list -> ActionsDeps.t -> Action.t list) =
  fun actions stbl -> 
    let visited_init = 
      List.fold_left (fun acc x -> MapAction.add x false acc) MapAction.empty actions 
    in
    let rec f (acc:Action.t list) (l:Action.t list) (stbl:t) (visited:bool MapAction.t) = 
      (* The graph contains no cycle! *)
      match l with
	     | [] -> List.rev acc
	     | x::tail -> 
	       if (try MapAction.find x visited with Not_found -> assert false)
	       then 
	         f acc tail stbl visited 
          else 
              let x_succ = ActionsDeps.find_deps stbl x in
              if x_succ = [] then
	             f (x::acc) tail stbl (MapAction.add x true visited)
              else
	             f acc (x_succ @ l) (ActionsDeps.remove_dep stbl x) visited
    in
    info "check_there_is_no_cycle...\n";
    check_there_is_no_cycle actions stbl;
    info "topo_sort...\n";
    f [] actions stbl visited_init

(*********************************************************************************)
(* From actions to gaos *)

let rec (gao_of_action: Action.t -> Soc.gao) = 
  fun (ck, il, ol, op, _lxm) ->
    (* nb : the list (encoded in a tree) is in the reverse order  *)
    let rec unpack_clock acc = function
      | Lic.BaseLic -> acc
      | Lic.ClockVar i -> acc (* should not occur? *)
      | Lic.On((value, cvar, _ctyp), outter_clock) ->
        let cc = Lv6Id.string_of_long2 value in        
        let acc = Soc.Case (cvar, [cc, [acc]]) in
        unpack_clock acc outter_clock
    in 
    unpack_clock (Soc.Call (ol, op, il)) ck

(*********************************************************************************)

(* In order to sort Soc.gao, I process in 3 stages ;
1. compute a total ordering of Actions.t according to the deps
2. transform Actions.t into Soc.gao (List.map)
3. factorize Soc.gao by looking at consecutive gao that have the same guard

*)

let (optimize_test_openning: Soc.gao list -> ActionsDeps.t -> Soc.gao list) =
  fun gaol deps ->
    let rec aux acc gaol = match gaol with 
      | []  -> List.rev acc
      | [a] -> List.rev (a::acc)
      | Soc.Call(o,op,i)::tail -> aux (Soc.Call(o,op,i)::acc) tail 
      | a1::Soc.Call(o,op,i)::tail ->  aux (Soc.Call(o,op,i)::a1::acc) tail 
      | Soc.Case(v1,l1)::Soc.Case(v2,l2)::tail -> 
        if v1 <> v2 then aux (Soc.Case(v1,l1)::acc) (Soc.Case(v2,l2)::tail) else
          let l = merge_gaol l1 l2 [] in
          aux acc (Soc.Case(v1,l)::tail) 
            
    and (merge_gaol : (string * Soc.gao list) list -> (string * Soc.gao list) list -> 
         (string * Soc.gao list) list -> (string * Soc.gao list) list) =
      fun l1 l2 acc -> 
        match l1 with
          | [] -> if l2 = [] then List.rev acc else List.rev_append acc l2
          | (x1,gaol1)::l1 -> 
            (match Lv6util.my_assoc x1 l2 with
              | None -> merge_gaol l1 l2 ((x1,gaol1)::acc)
              | Some(gaol2,l2) -> 
                let gaol = aux [] (gaol1@gaol2) in
                merge_gaol l1 l2 ((x1,gaol)::acc)
            )
    in
    aux [] gaol

(*********************************************************************************)
open Lv6MainArgs




let (f : Action.t list -> ActionsDeps.t -> Lxm.t -> Soc.gao list) = 
  fun actions deps lxm -> 
  (* =>  la liste d'actions en entrée contient des doublons !  *)
  try match global_opt.schedul_mode with
      | Simple -> (
        info "SortActions.f: topo_sort...\n";
        let actions = topo_sort actions deps in
        info "SortActions.f: gao_of_action...\n";
        let gaol = List.map gao_of_action actions in
        info "SortActions.f: optimize_test_openning actions...\n";
        optimize_test_openning gaol deps 
      )
      | Sort -> ( (* experimental scheduling *)
        let actions = List.sort SortActionsExpe.compare_actions actions in
        let actions = topo_sort actions deps in
        let gaol = List.map gao_of_action actions in
        optimize_test_openning gaol deps 
      )
      | Reorder -> ( (* experimental scheduling *)
        let actions = topo_sort actions deps in
        let actions = SortActionsExpe.group actions deps in 
        let gaol = List.map (List.map gao_of_action) actions in 
        SortActionsExpe.optimize_test_openning gaol 
      )
  with DependencyCycle(x,l) ->
    let l = List.map Action.to_string l in
    let msg = "A combinational cycle been detected "^
                (Lxm.details lxm)^" on \n  "^(Action.to_string x)^
                  "\n  "^(String.concat "\n  " l) ^ "\n\nHint: try to use --expand-nodes.\n"
    in
    raise (Lv6errors.Global_error msg)
